import inspect

from django.conf import settings
from django.test import TestCase
from django.utils.translation import activate
from django.db.models.query_utils import CollectedObjects

from easymode.tests.models import ManagerErrorModel
from easymode.tests.testcases import initdb
from easymode.tree import introspection
from easymode.i18n.meta import DefaultFieldDescriptor

__all__ = ('TestIntrospection',)

@initdb
class TestIntrospection(TestCase):
    """Tests for the introspection package in easymode"""
    
    fixtures = ['managererrormodel']
    
    def test_there_are_fixtures(self):
        """The fixtures hsould be instantiated coreectly"""
        self.failUnlessEqual(ManagerErrorModel.objects.count(), 2)
    
    def test_introspection_breaks(self):
        "inspect.getmembers should break in ManagerErrorModel"
        self.failUnlessRaises(AttributeError, inspect.getmembers, ManagerErrorModel)
    
    def test_bad_behaving_models_can_be_internationalized(self):
        """
        Models that throw exception when using module introspection,
        should be properly internationalized
        """
        obj = ManagerErrorModel.objects.get(pk=1)
        activate('en')
        self.failUnlessEqual(obj.a, 'english title a')
        self.failUnlessEqual(obj.b, 'uploads/11lekkerECHICKORIG.jpg')
        self.failUnlessEqual(obj.c, 'title c')
        activate('de')
        self.failUnlessEqual(obj.a, 'german titel a')
        self.failUnlessEqual(obj.b, 'uploads/111superlekkerechick.jpg')
        self.failUnlessEqual(obj.c, 'title c')
    
    def test_get_default_field_descriptors(self):
        "See if the DefaultFieldDescriptor is properly found"
        descriptors = introspection.get_default_field_descriptors(ManagerErrorModel)

        for (key, value) in descriptors:
            self.failUnlessEqual(type(value), DefaultFieldDescriptor)
        
        self.failUnlessEqual(len(descriptors), 2)

        descriptors = introspection.get_default_field_descriptors(ManagerErrorModel.objects.get(pk=1))

        for (key, value) in descriptors:
            self.failUnlessEqual(type(value), DefaultFieldDescriptor)
        
        self.failUnlessEqual(len(descriptors), 2)
        
        


"""
Contains widgets that can be used for admin models
with related items.
"""
from django import forms
from django.template.loader import render_to_string
from django.core import urlresolvers
from django.utils.encoding import force_unicode

from easymode.utils.languagecode import strip_language_code

class RenderLink(forms.Widget):
    """
    Renders a link to an admin page based on the primary key
    value of the model.
    """
    input_type = None
    
    def _has_changed(self, initial, data):
        return False
    def id_for_label(self, id_):
        return "hmm geen idee"
        
    def render(self, name, value, attrs=None):
        
        modelname = self.attrs['modelname']
        app_label = self.attrs['app_label']
        label = self.attrs['label']
        url_pattern = '%s:%s_%s_change' % ('admin', app_label, modelname)
        
        url = strip_language_code(urlresolvers.reverse(url_pattern, args=[value]))
            
        if value is None: value = ''
        final_attrs = self.build_attrs(attrs, type=self.input_type, name=name)
        if value != '':
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs['value'] = force_unicode(value)
        return render_to_string('tree/admin/widgets/foreignkeylink.html', locals())

class EmptyWidgetThatDoesNothing(forms.Widget):
    """does nothing"""
    def render(self, name, value, attrs=None):
        return render_to_string('tree/admin/widgets/foreignkeylink.html', locals())



#!/usr/bin/env python
# encoding: utf-8

"""
Module providing functions commonly used in shell scripting; requires
Python 2.5

The functions are as follows:

  - ffind(): finds files in a directory tree
  - ffindgrep(): finds files in a directory tree and matches their
    content to regular expressions
  - freplace(): in-place search/replace of files in a directory tree
    with regular expressions
  - printr(): prints the results of the ffind()/ffindgrep() functions

Please see the documentation strings of the particular functions for
detailed information.
"""

# Copyright: (c) 2007 Muharem Hrnjadovic
# created: 15/04/2007 09:31:25

__version__ = "$Id:$"
# $HeadURL $

import os, sys, types, re, functools

class ScriptError(Exception): pass

def ffind(path, namefs=None, relative=True):
    """
    Finds files in the directory tree starting at 'path' (filtered by the
    functions in the optional 'namefs' sequence); if the 'relative'
    flag is not set, the result sequence will contain absolute paths.

    Returns a sequence of paths for files found.
    """
    if not os.access(path, os.R_OK):
        raise ScriptError("cannot access path: '%s'" % path)

    fileList = [] # temporary files list
    try:
        for dir, subdirs, files in os.walk(path):
            fileList.extend('%s%s%s' % (dir, os.sep, f) for f in files)
        if not relative: fileList = map(os.path.abspath, fileList)
        if namefs:
            for ff in namefs: fileList = filter(ff, fileList)
    except Exception, e: raise ScriptError(str(e))
    return(fileList)

def ffindgrep(path, regexl, namefs=None, relative=True):
    """
    Finds files in the directory tree starting at 'path' (filtered by the
    functions in the optional 'namefs' sequence); if the 'relative'
    flag is not set, the result sequence will contain absolute paths.

    Additionaly, the file content will be filtered by the regular
    expressions in the 'regexl' sequence. Each entry in the latter
    is a
    
      - either a string (with the regex definition)
      - or a tuple with arguments accepted by re.compile() (the
        re.M and re.S flags will have no effect though)

    For all the files that pass the file name/content tests the function
    returns a dictionary where the

      - key is the file name and the
      - value is a string with lines filtered by 'regexl'
    """
    fileList = ffind(path, namefs=namefs, relative=relative)
    if not fileList: return dict()

    result = dict()

    try:
        # first compile the regular expressions
        ffuncs = []
        for redata in regexl:
            if type(redata) == types.StringType:
                ffuncs.append(re.compile(redata).search)
            elif type(redata) == types.TupleType:
                ffuncs.append(re.compile(*redata).search)
        # now grep in the files found
        for file in fileList:
            # read file content
            fhandle = open(file, 'r')
            fcontent = fhandle.read()
            fhandle.close()
            # split file content in lines
            lines = fcontent.splitlines()
            for ff in ffuncs:
                lines = filter(ff, lines)
                # there's no point in applying the remaining regular
                # expressions if we don't have any matching lines any more
                if not lines: break
            else:
                # the loop terminated normally; add this file to the
                # result set if there are any lines that matched
                if lines: result[file] = '\n'.join(map(str, lines))
    except Exception, e: raise ScriptError(str(e))
    return(result)

def freplace(path, regexl, namefs=None, bext='.bak'):
    """
    In-place search/replace of the content of all files (whose names
    pass the optional tests specified in 'namefs') in the directory
    tree starting with 'path' using the regular expressions in 'regexl'.

    Please note: 'regexl' is a sequence of 3-tuples, each having the
    following elements:

      - search string (Python regex syntax)
      - replace string (Python regex syntax)
      - regex flags or 'None' (re.compile syntax)

    Copies of the modified files are saved in backup files using the
    extension specified in 'bext'.

    The function returns the total number of files modified.
    """
    fileList = ffind(path, namefs=namefs)
    if not fileList: return 0

    filesChanged = 0
    cffl = []

    try:
        for searchs, replaces, reflags in regexl:
            if reflags is not None: regex = re.compile(searchs, reflags)
            else: regex = re.compile(searchs)
            # curry the regex.subn() function to supply the first
            # argument (the replacement string)
            cffl.append(functools.partial(regex.subn, replaces))
        for file in fileList:
            # read file content
            fhandle = open(file, 'r')
            text = fhandle.read()
            fhandle.close()
            substitutions = 0
            for cff in cffl:
                text, numOfChanges = cff(text)
                substitutions += numOfChanges
            if substitutions:
                # first move away the original file
                bakFileName = '%s%s' % (file, bext)
                if os.path.exists(bakFileName): os.unlink(bakFileName)
                os.rename(file, bakFileName)
                # now write the new file content
                fhandle = open(file, 'w')
                fhandle.write(text)
                fhandle.close()
                filesChanged += 1
    except Exception, e: raise ScriptError(str(e))

    # return the number of files that had some of their content changed
    return(filesChanged)

def printr(results):
    """
    prints the results of ffind()/ffindgrep() in a manner similar to
    the UNIX find utility
    """
    if type(results) == types.DictType:
        for f in sorted(results.keys()):
            sys.stdout.write("%s\n%s\n" % (results[f],f))
    else:
        for f in sorted(results):
            sys.stdout.write("%s\n" % f)

if __name__ == '__main__':
    pass




from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models

from easymode.admin.models.fields import DiocoreTextField, DiocoreHTMLField, FlashUrlField, DiocoreCharField
from easymode.i18n.decorators import I18n
from easymode.tree.decorators import toxml
from easymode.easypublisher.models import EasyPublisherModel


@toxml
@I18n('charfield')
class TestModel(models.Model):
    """test model"""

    charfield = models.CharField(max_length=255, unique=True)
    tags = models.ManyToManyField('TagModel', related_name='testmodels')
    
    def __unicode__(self):
        return u"%s%s - %s" % (self.__class__.__name__, self.pk, self.charfield)

    def natural_key(self):
        return (self.charfield,)
    

class TestSubModel(models.Model):
    """hangs onto testmodel"""

    testmodel = models.ForeignKey(TestModel, related_name='submodels')
    subcharfield = models.CharField(max_length=255)
    subintegerfield = models.IntegerField()

    def __unicode__(self):
        return u"%s%s - %s" % (self.__class__.__name__, self.pk, self.subcharfield)


class TestSecondSubmodel(models.Model):
    """hangs on to TestModel as well"""
    
    testmodel = models.ForeignKey(TestModel, related_name='secondsubmodels')
    ultrafield = models.TextField()
    
    def __unicode__(self):
        return self.ultrafield


class TestSubSubModel(models.Model):
    """Even subier sub model"""
    
    testsubmodel = models.ForeignKey(TestSubModel, related_name="subsubmodels")
    subsubcharfield = models.CharField(max_length=255)
    
    def __unicode__(self):
        return self.subsubcharfield


@I18n('value')
class TagModel(models.Model):
    
    value = models.CharField(max_length=23)


@toxml
@I18n('title', 'description')
class TestL10nModel(models.Model):

    title = DiocoreCharField(max_length=200, font="arial")
    description = DiocoreTextField(font="arial")
    body = DiocoreHTMLField(default='hi i am an error', font="arial")
    price = models.FloatField()

    def __unicode__(self):
        return u"%s%s" % (self.__class__.__name__, self.pk)


class GenericRelatedModel(models.Model):
    """it is used in a generic relation"""

    content_type = models.ForeignKey(ContentType)
    parent_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('content_type', 'parent_id')

    name = models.CharField(max_length=100)

    def __unicode__(self):
        return u"%s%s" % (self.__class__.__name__, self.pk)


@toxml
class TestGenericFkModel(models.Model):
    """it is used for testing"""

    name = models.CharField(max_length=100)
    relateds = generic.GenericRelation(GenericRelatedModel, content_type_field='content_type', object_id_field='parent_id')

    def __unicode__(self):
        return u"%s%s" % (self.__class__.__name__, self.pk)

class TestEasypublisherModel(EasyPublisherModel):
    label = models.CharField(max_length=200)
    nogeenlabel = models.CharField(max_length=200)
    def __unicode__(self):
        return "%s" % self.label
    
class TestEasypublisherRelatedModel(models.Model):
    parent = models.ForeignKey(TestEasypublisherModel)
    
    extra = models.CharField(max_length=200)
    def __unicode__(self):
        return "%s" % self.extra

@I18n('a', 'b')
class ManagerErrorModel(models.Model):
    a = models.CharField(max_length=255)
    b = models.ImageField(upload_to='uploads')
    c = models.TextField()

    def __unicode__(self):
        return u"%s %s %s %s" % (self.pk, self.a, self.b, self.c)

@I18n('number', 'another')
class FormTestModel(models.Model):
    number = models.IntegerField()
    other = models.CharField(max_length=255)
    another = models.CharField(max_length=255)
    


"""
These urls will be used when previewing a draft.
An extra query parameter will be passed to the view:

?preview=revision_id

Also the shortcut is not named 'r' but 'p' for preview.

include these urls in your urlconf like this::

    (r'^', include('easymode.easypublisher.urls')),

"""
from django.conf.urls.defaults import *

from easymode.utils.languagecode import get_language_codes_as_disjunction

languages = get_language_codes_as_disjunction()

urlpatterns = patterns('easymode.easypublisher.views',
    url(r'^admin/p/(?P<content_type_id>\d+)/(?P<object_id>.+)/(?P<revision_id>.+)/$', 'preview_draft', name='preview_draft'),
    (r'^(%(languages)s)/admin/p/(?P<content_type_id>\d+)/(?P<object_id>.+)/(?P<revision_id>.+)/$' % locals(), 'preview_draft'),
)



from django import template
from django.db.models import Q
from django.core.urlresolvers import reverse, NoReverseMatch
from django.utils.safestring import mark_safe

from reversion.models import Revision

from easymode.utils.languagecode import strip_language_code

register = template.Library()

#todo admin site is hardcoded, should get current admin name from somewhere
@register.tag
def draft_list_items(parser, token):
    """displays a list of all drafts, without the enclosing <ul>"""
    return DraftListItems()

class DraftListItems(template.Node):
    def render(self, context):
        request = context['request']

        #select revisions
        revisions = Revision.objects.filter(
            easypublishermetadata__status='draft',
            easypublishermetadata__language=request.LANGUAGE_CODE,
        ).select_related().distinct()
        
        output = u''
        for revision in revisions:
            # select version and make list item
            # since we are not saving related items, their object_id is set to u'None'
            # we don't need those.
            urls = {}
            versions = revision.version_set.filter(~Q(object_id='None'),).order_by('revision')
            for version in versions:
                opts = version.content_type.model_class()._meta
                info = opts.app_label, opts.module_name
                name = 'admin:%s_%s_draft' % info
                try:
                    url = reverse(name, args=[version.object_id, revision.id])
                    url = strip_language_code(url)
                    if not url in urls:
                        urls[url] = version
                except NoReverseMatch:
                    pass
            
            for (url, version) in urls.iteritems():
                output += u"<li><a href=\"%s\">%s</a></li>" % (url, version.object_repr)
    
        return mark_safe(output)



import copy

from django.db.models.base import ModelBase
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.contenttypes.generic import generic_inlineformset_factory
from django.contrib.contenttypes.generic import BaseGenericInlineFormSet
from django.contrib.admin.util import flatten_fieldsets

from easymode.i18n.admin import forms
from easymode.i18n.admin.generic import LocalizableGenericInlineFormSet
from easymode.utils.languagecode import get_all_language_codes, localize_fieldnames

__all__ = ('L10n', 'lazy_localized_list')

def compute_prohibited(fields, exclude, localized):
    return set(fields) - set(exclude) - set(localized)

class lazy_localized_list(list):
    """
    A descriptor that can get passed contrib.admin.validation.check_isseq
    undetected. It will give the 'real' name of an internationalized
    property when localized.
    """

    def __new__(cls, sequence, localized_fieldnames):
        if type(sequence) is lazy_localized_list:
            return sequence
        return list.__new__(cls, sequence)
        
    def __init__(self, sequence, localized_fieldnames):
        self.localized_fieldnames = localized_fieldnames
        super(lazy_localized_list, self).__init__(sequence)
    
    def __get__(self, obj, typ=None):
        """
        returns a localized version of the list this descriptor
        was initialized with.
        """
        return localize_fieldnames(self, self.localized_fieldnames)

class L10n(object):
    """
    Localise the admin class.
    
    filter fields to be displayed on a modeladmin class.
    All extra fields generated by I18n decorator should be hidden.
    Also uses a special form class that shows the field in the
    current language.
    
    This decorator also adds the 'can_edit_global_fields' permission to
    the model, only users that have this permission can actually edit fields
    that are not localised.
    
    >>> from easymode.i18n.admin import L10n
    >>> @L10n(SomeModel)
    >>> class SomeModelAdmin(admin.ModelAdmin):
    >>>     pass
    
    or you can also:
    
    >>> admin.site.register(SomeModel,L10n(SomeModel, admin.ModelAdmin))
    
    This is usefull if you are using the same admin class for multiple models.
    
    If you have defined the model property on your admin class you can just
    leave it at:
    
    >>> @L10n
    >>> class SomeModelAdminWithModelDefined(admin.ModelAdmin):
    >>>     model = SomeModel
    """
    
    error_no_model = "L10n: %s does not have model defined, but no model was passed to L10n either"
    
    def __new__(typ, model_or_admin=None, cls=None, attrs=None):
        """
        Construct object and if cls is passed as well apply the object to that 
        immediately. This makes this decorator a factory as well.
        """
        obj = object.__new__(typ)
        
        if isinstance(model_or_admin, ModelBase):
            # if model_or_admin is a model class set the model.
            obj.model = model_or_admin
        elif issubclass(model_or_admin, BaseModelAdmin):
            # if model_or_admin is an admin class, model MUST be defined on that class.
            if hasattr(model_or_admin, 'model'):
                obj.model = model_or_admin.model
                # set cls to the admin class
                cls = model_or_admin
            else:
                raise(AttributeError(L10n.error_no_model % model_or_admin.__name__))
        else:
            raise TypeError("L10n can not accept paramters of type %s" % model_or_admin.__name__)
            
        # if cls is defined call __call__
        if cls:                    
            # first make a new class so the decorator does not mess up the
            # class passed. This new class is then modified by the decorator.
            descendant = type(obj.model.__name__ + cls.__name__, (cls,), attrs or {'model':obj.model})
            return obj.__call__(descendant)
            
        return obj
    
    def __call__(self, cls):
        """run the filter on the class to be decorated"""
        if hasattr(cls, 'model'):
            self.model = cls.model
        elif not self.model:
            raise AttributeError(L10n.error_no_model % (cls.__name__) )

        # gather names of fields added by I18n
        added_fields = []
        for field in self.model.localized_fields:
            for language in get_all_language_codes():
                added_fields.append("%s_%s" % (field, language))

        # hide added fields from form and admin
        cls.exclude = added_fields
        cls.form = forms.make_localised_form(self.model, cls.form, exclude=added_fields)

        # determine name of the permission to edit untranslated fields
        permisson_name = "%s.can_edit_untranslated_fields_of_%s" % (
            self.model._meta.app_label, self.model.__name__.lower()
        )
        
        # make sure that fields become read_only when no permission is given.
        def get_readonly_fields(self, request, obj=None):
            if not request.user.has_perm(permisson_name):
                fields = self.fields or map(lambda x: x.name, self.model._meta.fields)
                # remove primary key because we don't show that, not even uneditable
                fields.pop(self.model._meta.pk_index())
                prohibited_fields = compute_prohibited(fields, self.exclude, self.model.localized_fields)
                
                return set(self.readonly_fields).union(prohibited_fields)
            
            return self.readonly_fields
        
        cls.get_readonly_fields = get_readonly_fields
        
        # override some views to hide fields which are not localized
        if hasattr(cls, 'change_view'):
            # BaseModelAdmin.__init__ will mess up our lazy lists if the following is
            # not allready defined
            if 'action_checkbox' not in cls.list_display and cls.actions is not None:
                cls.list_display = ['action_checkbox'] +  list(cls.list_display)
            
            if not cls.list_display_links:
                for name in cls.list_display:
                    if name != 'action_checkbox':
                        cls.list_display_links = [name]
                        break

            # Make certain properties lazy and internationalized
            cls.list_display_links = lazy_localized_list(cls.list_display_links, self.model.localized_fields)
            cls.list_display = lazy_localized_list(cls.list_display, self.model.localized_fields)
            cls.list_editable = lazy_localized_list(cls.list_editable, self.model.localized_fields)
            cls.search_fields = lazy_localized_list(cls.search_fields, self.model.localized_fields)
            
        else:
            def get_formset(self, request, obj=None, **kwargs):
                if self.declared_fieldsets:
                    fields = flatten_fieldsets(self.declared_fieldsets)
                else:
                    fields = None
                if self.exclude is None:
                    exclude = []
                else:
                    exclude = list(self.exclude)
                exclude.extend(self.get_readonly_fields(request, obj))
                exclude = exclude or None

                if not hasattr(self, 'ct_fk_field'):
                    return super(cls, self).get_formset(request, obj, **kwargs)
                else:
                    # TODO:
                    # this code can be deleted if django fixes GenericInlineModelAdmin it's
                    # get_formset signature so it looks like InlineModelAdmin
                    defaults = {
                        "ct_field": self.ct_field,
                        "fk_field": self.ct_fk_field,
                        "form": self.form,
                        "formfield_callback": self.formfield_for_dbfield,
                        "formset": self.formset,
                        "extra": self.extra,
                        "can_delete": self.can_delete,
                        "can_order": False,
                        "fields": fields,
                        "max_num": self.max_num,
                        "exclude": exclude
                    }
                    
                    defaults.update(kwargs)
                    # the BaseGenericInlineFormSet does not work too well
                    # with modified models, so use LocalizableGenericInlineFormSet.
                    if self.formset is BaseGenericInlineFormSet \
                        or self.formset.__class__ is BaseGenericInlineFormSet:
                        defaults['formset'] = LocalizableGenericInlineFormSet
                        
                    return generic_inlineformset_factory(self.model, **defaults)
            
            cls.get_formset = get_formset
        
            
        return cls



"""
    easy_copy_language <source locale> <target locale> <app>
    
    will copy all internationalised fields from source to target for the app in question.
    
    example:
    
    ./manage.py easy_copy_language en de myapp.mymodel
    
    will copy the en fields into the de locale for myapp.mymodel
"""

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import get_models, get_app, get_model
from django.utils.encoding import force_unicode
from django.utils import translation

class Command(BaseCommand):
    help = __doc__
    
    def handle(self, source, target, app=None, **options):
        """ command execution """
        translation.activate(settings.LANGUAGE_CODE)
        if app:
            unpack = app.split('.')

            if len(unpack) == 2:
                models = [get_model(unpack[0], unpack[1])]
            elif len(unpack) == 1:
                models = get_models(get_app(unpack[0]))
        else:
            models = get_models()
            
        for model in models:
            if hasattr(model, 'localized_fields'):

                model_full_name = '%s.%s' % (model._meta.app_label, model._meta.module_name)
                update_instances = set()
                messages = []
                
                for instance in model.objects.all():
                    for field in model.localized_fields:
                        source_field = "%s_%s" % (field, source)
                        target_field = "%s_%s" % (field, target)
                        
                        if  hasattr(instance, source_field) and hasattr(instance, target_field):
                            source_field_value = getattr(instance, source_field)
                            target_field_value = getattr(instance, target_field)

                            if target_field_value in (None, u'')\
                                and source_field_value not in (None, u''):
                                setattr(instance, target_field, force_unicode(source_field_value))
                                update_instances.add(instance)
                                messages.append(u"%s %s %s will become %s" % (model_full_name, instance, target_field, force_unicode(source_field_value)))

                if len(update_instances):
                    if self.ask_for_confirmation(messages, u'%s.%s' % (model._meta.app_label, model._meta.module_name)):
                        for update_instance in update_instances:
                            print u"saving %s" % update_instance
                            update_instance.save()
                
                
    def ask_for_confirmation(self, messages, model_full_name):
        print u'\nData to update "%s":' % model_full_name
        for message in messages:
            print u'   %s' % message
        while True:
            prompt = u'\nAre you sure that you want to update %s: (y/n) [n]: ' % model_full_name
            answer = raw_input(prompt).strip()
            if answer == '':
                return False
            elif answer not in ('y', 'n', 'yes', 'no'):
                print u'Please answer yes or no'
            elif answer == 'y' or answer == 'yes':
                return True
            else:
                return False
            
    


"""
Tools for debugging python applications
"""
from StringIO import StringIO
import inspect

__all__ = ('stack_trace', 'middleware')

def stack_trace(depth=None):
    """
    returns a print friendly stack trace at the current frame,
    without aborting the application.
    
    :param depth: The depth of the stack trace. if omitted, the entire
        stack will be printed.
    
    usage::
    
        print stack_trace(10)
    """
    frames = inspect.stack()[2:]
    if depth:
        frames = frames[:depth]
        
    result = StringIO()
    result.write("----------------------------------------------------\n")
    for (frame, file, line, context, code, status) in frames:
        result.write("In %s from %s\n%s %s" % (context, file, line, "\n".join(code)))
        result.write("----------------------------------------------------\n")        
    return result.getvalue()



"""
Imports all testcases found in the 'testcases' subfolder of 
folder containg the file parameter.
Doing that is enough to make the testrunner find them.
"""
import glob
import os
import re
import sys

from django.conf import settings

from easymode.easypublisher.utils import filter_unpublished
from easymode.utils import first_match, bases_walker, url_add_params
from easymode.utils.languagecode import get_language_codes,\
    get_language_codes_as_disjunction, get_language_code_from_shorthand,\
    localize_fieldnames, get_real_fieldname, strip_language_code,\
    get_short_language_codes
from easymode.utils.standin import standin_for

# check if some required settings are fulfilled
if 'de' not in get_language_codes():
    raise Exception('the language "de" must be in your LANGUAGES to run the test suite')
if 'en-us' not in get_language_codes():
    raise Exception('the language "en-us" must be in your LANGUAGES to run the test suite')

if settings.LANGUAGE_CODE is not 'en':
    raise Exception('To run the test suite the LANGUAGE_CODE must be set to "en"')

# find all tests in the testcases folder
path = os.path.join(os.path.dirname(__file__), 'testcases')
pattern = os.path.join(path , r'(.+).py')

filenames = glob.glob(path + '/*.py')

if len(filenames): # delete __init__.py. yes it is always first.
    del(filenames[0])

# import all the testcases in the current scope
for filename in filenames:
    modulename = re.sub(pattern, r'\1', filename)
    try:
       testcase_module = __import__("easymode.tests.testcases.%s" % modulename, locals(), globals(), [''])
       test_class_names = getattr(testcase_module, '__all__')
       for test in testcase_module.__all__:
           setattr(sys.modules[__name__], test, getattr(testcase_module, test))
    except Exception, e:
       print "%s in %s, please correct if you want to run this test suite." % (e, modulename)


# doctests
__test__ = {
    "test_get_language_codes" : get_language_codes,
    "test_get_short_language_codes" : get_short_language_codes,
    "test_get_language_code_from_shorthand" : get_language_code_from_shorthand,
    "test_strip_language_code" : strip_language_code,
    "test_get_real_fieldname" : get_real_fieldname,
    "test_localize_fieldnames" : localize_fieldnames,
    "test_get_language_codes_as_disjunction" : get_language_codes_as_disjunction,
    "test_first_match" : first_match,
    "test_bases_walker" : bases_walker,
    "test_standin_for": standin_for,
    'test_url_add_params' : url_add_params,
    'test_filter_unpublished': filter_unpublished,
}



# -*- coding: utf-8 -*-
import warnings
import re
import shutil
import os
import errno
import gc
import time

import gettext as gettext_module
from hashlib import sha1

from os.path import join, isdir


from django.test import TestCase
from django.db import IntegrityError
from django.db.models.signals import post_save
from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils import translation
from django.utils.translation import trans_real

if 'rosetta' in settings.INSTALLED_APPS:
    from rosetta import polib
else:
    import polib

from easymode.utils import first_match, mutex, SemaphoreException
from easymode.utils.languagecode import get_language_codes
from easymode.i18n import meta
from easymode.tests.testcases import initdb
from easymode.tests import models
from easymode.i18n.gettext import MakeModelMessages
from easymode.i18n import gettext
from easymode.tests.testutils.scriptutil import ffindgrep
from easymode import tree


__all__ = ('Testi18n',)

@initdb
class Testi18n(TestCase):
    """tests for the internationalisation/localisation support"""

    fixtures = ['auth-user', 'auth-group']
    
    def setUp(self):
        self.settingsManager.set(
            MSGID_LANUAGE='en',
            FALLBACK_LANGUAGES={'de': ['en-us']}
        )
        gc.collect()
        translation.activate(settings.LANGUAGE_CODE)
        
        try:
            t = models.TestModel(charfield='Hoi Ik ben de root node')
            t.save()
            f = t.submodels.create(subcharfield="Hoi Ik ben de first level sub node", subintegerfield=10)
            s = t.submodels.create(subcharfield="Hoi ik ben de tweede first level sub node", subintegerfield=100)
            w = t.secondsubmodels.create(ultrafield="Sed tempor. Ut felis. Maecenas erat.")
            f.subsubmodels.create(subsubcharfield="Hoi ik ben de thord level sub node")
            f.subsubmodels.create(subsubcharfield="Hoi ik ben de third level sub node")
            s.subsubmodels.create(subsubcharfield="Hoi ik ben de third level sub node")
            s.subsubmodels.create(subsubcharfield="Hoi ik ben de third level sub node")        
            t.save()          
        except OSError as e:
            if e.errno == 24:
                print "hai er zit een bug in de mutex"

    # some vars
    title_en = 'HI I GUESS I AM THE TITLE IN ENGLISH'
    title_de = 'ICH LIEBE DICH'

    arr = [None, None, None, 'hai', 'True', True, 'last']
    
    def setup_l10n_model(self):
        translation.activate('en')
        i = models.TestL10nModel(title="Ik ben de groot moeftie van cambodja", 
            description="Ik weet weinig tot niks van brei cursussen", 
            body="Het zou kunnen dat linksaf slaan hier een goede keuze is.",
            price=12.0,
            )
        i.save()
        return i
    
    def setup_mofile_with_entry(self, poentry, language='en'):
        locale_location = join(settings.LOCALE_DIR, 'locale', language, 'LC_MESSAGES')
        pofile = polib.pofile(join(locale_location, 'django.po'))

        pofile.append(poentry)
        pofile.save()
        pofile.save_as_mofile(join(locale_location, 'django.mo'))
        
        jit_locale = gettext_module.translation('django', join(settings.LOCALE_DIR, 'locale'), [trans_real.to_locale(language)], trans_real.DjangoTranslation)
        jit_locale.set_language(language)
        
        locale = trans_real.translation(language)
        locale.merge(jit_locale)
        
    
    def test_locale_dir_is_not_project_dir(self):
        """The locale dir should be somewhere safe"""
        assert(not (os.path.normpath(settings.LOCALE_DIR) == os.path.normpath(settings.PROJECT_DIR)))
        
    def test_save_signal_is_caught(self):
        """when a model is saved a signal should be caught"""

        def stub(sender, **kwargs):
            self.save_was_caught = True
            self.save_was_a_created = kwargs['created']

        self.save_was_caught = False
            
        post_save.connect(stub, models.TestModel)
        i = models.TestModel.objects.get(pk=1)
        i.charfield = 'hai mom'
        i.save()
        assert(self.save_was_caught)
        
    def test_created_property_is_set(self):
        """A signal sent when a model is created should have the created property set to true"""
        def stub(sender, **kwargs):
            self.save_was_caught = True
            self.save_was_a_created = kwargs['created']

        self.save_was_caught = False
        self.save_was_a_created = False
        
        post_save.connect(stub, models.TestModel)
        i = models.TestModel(charfield='hai mother')
        i.save()
        assert(self.save_was_caught)
        assert(self.save_was_a_created)

    def test_we_have_the_proper_version_of_gettext(self):
        """The proper version of xgettext (>= 0.15) should be installed on the server."""
        
        if gettext.XGETTEXT_REENCODES_UTF8 is True:
            warnings.warn('while not fatal it is better to install a version of gettext \
                higher or equal to 0.15')
                    
    def test_model_can_be_serialized(self):
        """A model should be serializable to po"""
        db_gettext = MakeModelMessages()
        i = models.TestModel.objects.get(pk=1)

        m = str(db_gettext.poify(i))

        assert(m.index('msgid "Hoi Ik ben de root node"') is not None)
    
    def test_model_to_po(self):
        """A newly created model should be serialized to po"""
        locale_dir = join(settings.LOCALE_DIR, 'locale')
        if isdir(locale_dir):
            shutil.rmtree(locale_dir)
            
        assert(not isdir(locale_dir))
        i = models.TestModel(charfield='ik word serialized to po')
        i.save()
        assert(isdir(locale_dir))
        
        result = ffindgrep(locale_dir, ['^msgid "ik word serialized to po"$'])
        assert(result != {})
        
    def test_meta(self):
        """field should have a title_en as well as a title if default locale is en"""
        i = self.setup_l10n_model()
        
        assert(i.title_en == i.title == 'Ik ben de groot moeftie van cambodja')
    
    def test_first_true(self):

        assert(first_match(lambda x: x or None, self.arr))

    def test_meta_field_finder(self):
        """The correct localised version of the title property should be found"""
        field = meta.get_localized_property(self, 'title')
        assert(field == self.title_en)
        translation.activate('de')
        field = meta.get_localized_property(self, 'title')
        assert(field == self.title_de)

    def test_meta_get_localized_field_name(self):
        """get_localized_field_name should return the correct localized property name"""
        field = meta.get_localized_field_name(self, 'title')
        assert(field == 'title_en')
            
    def test_meta_now_selects_correct_field_on_propert_write(self):
        """Easymode should write to title_en if language is en and property is title"""            
        i = self.setup_l10n_model()
        i.title = 'ik wijzig em ff ok?'
        i.save()
        assert(i.title == i.title_en == 'ik wijzig em ff ok?')
    
    def test_ugettext_lazy(self):
        """ugettext_lazy should return translated strings, if they are in the locale"""

        self.setup_l10n_model()
        
        poentry = polib.POEntry(msgid='test string', msgstr='LOL DID YOU REALLY THINK I WAS NUB?')
        
        self.setup_mofile_with_entry(poentry)
        
        lol = translation.ugettext_lazy("test string")
        assert(unicode(lol) == u'LOL DID YOU REALLY THINK I WAS NUB?')
    
    def test_adding_l10n_decorator_means_model_gets_rendered_to_po_file(self):
        """If a model is created, which is decorated with l10n, it should be added to the po file"""
        self.setup_l10n_model()

        result = ffindgrep(join(settings.LOCALE_DIR, 'locale'), 
            ['^msgid "Ik ben de groot moeftie van cambodja"$'])
        assert(result != {})
        
    def test_nulled_translations_get_translated_from_po_file(self):
        """If a property which is a translation is accessed but it is None, the translation from the po file should be used"""
        i = self.setup_l10n_model()
        
        assert(i.title == i.title_en == u'Ik ben de groot moeftie van cambodja')
        
        poentry = polib.POEntry(msgid=u'Ik ben de groot moeftie van cambodja', msgstr=u'Ik ben een zwerver')        
        self.setup_mofile_with_entry(poentry, 'de')
        
        translation.activate('de')

        assert(i.title_de is None)
        assert(i.title == u'Ik ben een zwerver')
    
    def test_translations_that_are_not_null_should_not_come_from_the_po_file(self):
        """If a translation is not null, that value should be returned."""
        i = self.setup_l10n_model()
        i.title_de = u'Ik ben een konijntje'
        i.save()
        
        assert(i.title == i.title_en == u'Ik ben de groot moeftie van cambodja')
        poentry = polib.POEntry(msgid=u'Ik ben de groot moeftie van cambodja', msgstr=u'Ik ben een zwerver')        
        self.setup_mofile_with_entry(poentry, 'de')
        
        translation.activate('de')

        assert(i.title_de is not None)
        assert(not (i.title == u'Ik ben een zwerver'))
        assert(i.title == u'Ik ben een konijntje')
        
    def test_fallback_translation(self):
        """The fallback from the database should win from the gettext fallback"""
        translation.activate(settings.LANGUAGE_CODE)
        value = 'Hoi Ik ben de root node'
        t = models.TestModel.objects.get(**{'charfield_en': value})

        # test gettext fallback
        translation.activate('de')
        self.assertEqual(t.charfield, 'Hoi Ik ben de root node')

        # change the translation in the fallback language of 'de'
        translation.activate('en-us')
        t.charfield = 'Hoi I am An american'
        t.save()
        self.assertEqual(t.charfield, 'Hoi I am An american')      
        
        # test database fallback
        translation.activate('de')
        self.assertEqual(t.charfield, 'Hoi I am An american')


    def test_translation_nomsgid(self):
        """
        When the msgid is None or empty, the value from the fallback language
        should be used.
        """
        
        # Extra setup
        translation.activate('en-us')
        i = models.TestModel(charfield='Woot, not failed!')
        i.save()
        self.assertTrue(i.charfield == getattr(i, 'charfield_en-us') == 'Woot, not failed!')
        translation.activate('de')

        # Test
        self.assertEqual(i.charfield, 'Woot, not failed!')


    def test_translated_fields_handle_correctly_under_to_xml(self):
        """A field that is translated should show the correct value when converted to xml"""

        i = self.setup_l10n_model()

        poentry = polib.POEntry(msgid=u'Ik ben de groot moeftie van cambodja', msgstr=u'Ik ben een zwerver')        
        self.setup_mofile_with_entry(poentry, 'de')

        xml_representation = tree.xml(i)

        # todo:
        # DefaultFieldDescriptor is now serialized as a charfield type.
        # add the different serializable types to the xslt and the test should be able
        # to verify the correct workings then.
        # contains_default_field_descriptor = re.search(r'DefaultFieldDescriptor', xml_representation)
        # self.assertTrue(contains_default_field_descriptor)
        
        contains_moeftie = re.search(r'Ik ben de groot moeftie van cambodja', xml_representation)
        contains_konijntje = contains_konijntje = re.search(r'Ik ben een konijntje', xml_representation)        
        self.assertTrue(contains_moeftie)
        self.assertFalse(contains_konijntje)

        translation.activate('de')
        i.title = u'Ik ben een duits konijntje'
        xml_representation = tree.xml(i)
        contains_konijntje = re.search(r'Ik ben een duits konijntje', xml_representation)
        self.assertTrue(contains_konijntje)

        i.title_de = u'Ik ben geen konijntje'
        i.save()

        xml_representation = tree.xml(i)
        contains_konijntje = re.search(r'Ik ben geen konijntje', xml_representation)
        self.assertTrue(contains_konijntje)

                
    def test_permissions_for_change_view_are_evaluated_each_request(self):
        "The permissions on different views should be evaluated each request"

        post_data = {
            'price': 100,
            'title': 'Your mother',
            'description': 'lmao',
            'body': 'MY BAR IS LOL HAHA',
        }

        admin_add = reverse('admin:tests_testl10nmodel_add')
        admin_change = reverse('admin:tests_testl10nmodel_change', args=[1])

        try:
            can_login = self.client.login(username='admin', password='admin')
            self.client.post(admin_add, post_data)

            can_login = self.client.login(username='editor', password='editor')
            response = self.client.post(admin_change, post_data)
            self.failUnlessEqual(response.status_code, 302)

            can_login = self.client.login(username='admin', password='admin')
            self.client.post(admin_add, post_data)
        except IntegrityError as e:
            self.fail(e)

        self.failUnlessEqual(can_login, True)
            



"""
Contains tools to enable preview of drafts.
"""
from lxml import etree
from reversion.models import Revision
from copy import deepcopy
from easymode.tree.serializers import RecursiveXmlSerializer


__all__ = ('filter_unpublished', 'insert_draft')

class RevisionTree(object):
    """
    This class builds an xml tree from a revision, just
    like :class:`~easymode.tree.serializers.RecursiveXmlSerializer` does
    from a regular model.
    
    :attr topnode: The main node in the revision, with all child nodes nested inside.
    :attr nodelist: All nodes in the revision in a flat list.
    :meth xpath_for_node: Return an xpath statement which can be used to look
        in another xml document for the same node.
    """
    
    def __init__(self, revision_id):
        nodelist = []

        revision = Revision.objects.get(pk=revision_id)
        serializer = RecursiveXmlSerializer()
        
        # turn all version objects into lxml elements
        for ver in revision.version_set.all():
            model_instance = ver.get_object_version().object
            draft_xml = serializer.serialize([model_instance])
            draft_doc = etree.fromstring(draft_xml)
            draft_nodes = draft_doc.xpath(r'/django-objects/object')
            nodelist += draft_nodes
        
        topnode = self._find_topnode(nodelist)
        nodelist.remove(topnode)

        self.nodelist = nodelist
        # create xml topnode with nested items
        self._append_children(topnode, nodelist)
        self.topnode = topnode

    def xpath_for_node(self, node):
        return r"//object[@pk='%(pk)s' and @model='%(model)s']" \
            % {
            'pk': node.get('pk'),
            'model': node.get('model'),
        }
        
    def _get_items_by_model_and_pk(self,nodelist, model, pk):
        object_filter = lambda x: x.get('model') == model and x.get('pk') == pk
        return filter(object_filter, nodelist)

    def _get_relations(self, node):
        return map(lambda x: (x.get('to'), x.text), 
            node.xpath(r"field[@to and text() != 'None']"))

    def _get_parents(self, node, nodelist):
        result = []
        for (model, pk) in self._get_relations(node):
            result += self._get_items_by_model_and_pk(nodelist, model, pk)
        
        return result

    def _get_children(self, node, nodelist):
        children = []
        pk = node.get('pk')
        model = node.get('model')
        
        return [childnode for childnode in nodelist if \
            (model, pk) in self._get_relations(childnode)]

    def _find_topnode(self, nodelist):
        for node in nodelist:
            parents = self._get_parents(node, nodelist)
            if len(parents) == 0:
                return node
            else:
                return self._find_topnode(parents)
                
        return None

    def _append_children(self, node, nodelist):
        for child in self._get_children(node, nodelist):
            self._append_children(child, nodelist)
            node.append(child)


def insert_draft(revision_id, xml):
    """
    Insert the data belonging to revision revision_id into the xml
    
    :param revision_id: The id of the :class:`~reversion.models.Revision` we want to include.
    :param xml: A string that can be parsed as valid xml.
    :result: An xml string with the revision data included.
    """
    
    revision_tree = RevisionTree(revision_id)
    xml_doc = etree.fromstring(xml)
    
    # each separated versioned object should be replaced separately, because
    # for example list_editable view changes have many unrelated objects in 
    # a single revision
    for draft_node in revision_tree.nodelist:
        for matching_node in xml_doc.xpath(revision_tree.xpath_for_node(draft_node)):
            matching_node.getparent().replace(matching_node, deepcopy(draft_node))
    
    # when the revision is normal and not from a list_editable view, there is one
    # top node and the rest are related items. Create a tree structure with all
    # the related items as a tree, like easymode's recursive xml serializer does
    # as well.
    top_node = revision_tree.topnode
    
    for matching_node in xml_doc.xpath(revision_tree.xpath_for_node(top_node)):
        matching_node.getparent().replace(matching_node, deepcopy(top_node))
    
    return etree.tostring(xml_doc)


def filter_unpublished(xml):
    """
    Filter all unpublished objects from the xml.
    
    >>> a = '''<root>
    ... <object><field name="published">False</field>hahaha</object>
    ... <object><field name="published">True</field>hihih</object>
    ... <object><field name="koe">False</field>lololol</object>
    ... <object>
    ...     <field name="published">False</field>
    ...     hahaha
    ...     <object><field name="koe">False</field>doe eens niet</object>
    ... </object></root>'''
    >>>
    >>> filter_unpublished(a)
    '<root>\\n<object><field name="published">True</field>hihih</object>\\n<object><field name="koe">False</field>lololol</object>\\n</root>'
    
    :param xml: A string that can be parsed as valid xml.
    :result: An xml string with only published models.
    """

    select_unpublished = r"//object[field[@name='published' and text() = 'False']]"
    xml_tree = etree.fromstring(xml)
    for node in xml_tree.xpath(select_unpublished):
        node.getparent().remove(node)

    return etree.tostring(xml_tree)


# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding field 'NotABase.kak'
        db.add_column('foobar_notabase', 'kak', self.gf('django.db.models.fields.related.OneToOneField')(default=1, related_name='kaks', unique=True, to=orm['foobar.Baz']), keep_default=False)


    def backwards(self, orm):
        
        # Deleting field 'NotABase.kak'
        db.delete_column('foobar_notabase', 'kak_id')


    models = {
        'foobar.bar': {
            'Meta': {'object_name': 'Bar'},
            'date': ('django.db.models.fields.DateField', [], {}),
            'foo': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'bars'", 'to': "orm['foobar.Foo']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'foobar.baz': {
            'Meta': {'object_name': 'Baz'},
            'baz': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'independancy_level': ('django.db.models.fields.IntegerField', [], {}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'foobar.foo': {
            'Meta': {'object_name': 'Foo'},
            'address': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'bar': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'barstool': ('django.db.models.fields.TextField', [], {'max_length': '4'}),
            'city_de': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'city_en': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True'}),
            'city_en-us': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'date_de': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'date_en': ('django.db.models.fields.DateField', [], {'null': 'True'}),
            'date_en-us': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'website': ('django.db.models.fields.URLField', [], {'max_length': '200'})
        },
        'foobar.notabase': {
            'Meta': {'object_name': 'NotABase'},
            'extra': ('django.db.models.fields.CharField', [], {'max_length': '344'}),
            'hitehre': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['foobar.Baz']", 'unique': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'kak': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'kaks'", 'unique': 'True', 'to': "orm['foobar.Baz']"}),
            'situp': ('django.db.models.fields.IntegerField', [], {})
        }
    }

    complete_apps = ['foobar']



"""
:func:`~django.shortcuts.render_to_response` like functions, that help you
create views with filtered xml.
"""
from django.http import HttpResponse

from easymode.easypublisher.utils import filter_unpublished, insert_draft
from easymode.tree import xml as to_xml
from easymode.xslt.response import render_xml_to_string


def render_to_response_with_revision(template, queryset, revision_id, params=None, mimetype='text/html'):
    """
    Renders a queryset to xml, but inserts the data from revision *revision_id*
    in the appropriate place. This can be used to enable previews
    
    :param template: The xslt template path.
    :param queryset: The queryset that should be turned into xml.
    :param revision_id: The id of the revision whose data we want to include in the xml.
    :param params: xslt params, correctly escaped using :func:`~easymode.xslt.response.prepare_string_param`
    :param mimetype: The mimetype of the response.
    :result: A :class:`~django.http.HttpResponse` with xml, including the data in revision *revision_id*
    """
    xml = insert_draft(revision_id, to_xml(queryset))
    return HttpResponse(render_xml_to_string(template, xml, params), mimetype=mimetype)

def render_to_response_filtered(template, queryset, params=None, mimetype='text/xml'):
    """
    Renders a queryset to xml but it strips all objects that have a field named
    published that is set to False.
    
    :param template: The xslt template path.
    :param queryset: The queryset that should be turned into xml.
    :param params: xslt params, correctly escaped using :func:`~easymode.xslt.response.prepare_string_param`
    :param mimetype: The mimetype of the response.
    :result: A :class:`~django.http.HttpResponse` with xml, with only published models.
    """
    xml = filter_unpublished(to_xml(queryset))
    return HttpResponse(render_xml_to_string(template, xml, params), mimetype=mimetype)


# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'NotABase'
        db.create_table('foobar_notabase', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('extra', self.gf('django.db.models.fields.CharField')(max_length=344)),
            ('situp', self.gf('django.db.models.fields.IntegerField')()),
            ('hitehre', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['foobar.Baz'], unique=True)),
        ))
        db.send_create_signal('foobar', ['NotABase'])


    def backwards(self, orm):
        
        # Deleting model 'NotABase'
        db.delete_table('foobar_notabase')


    models = {
        'foobar.bar': {
            'Meta': {'object_name': 'Bar'},
            'date': ('django.db.models.fields.DateField', [], {}),
            'foo': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'bars'", 'to': "orm['foobar.Foo']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'foobar.baz': {
            'Meta': {'object_name': 'Baz'},
            'baz': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'independancy_level': ('django.db.models.fields.IntegerField', [], {}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'foobar.foo': {
            'Meta': {'object_name': 'Foo'},
            'address': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'bar': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'barstool': ('django.db.models.fields.TextField', [], {'max_length': '4'}),
            'city_de': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'city_en': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True'}),
            'city_en-us': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'date_de': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'date_en': ('django.db.models.fields.DateField', [], {'null': 'True'}),
            'date_en-us': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'website': ('django.db.models.fields.URLField', [], {'max_length': '200'})
        },
        'foobar.notabase': {
            'Meta': {'object_name': 'NotABase'},
            'extra': ('django.db.models.fields.CharField', [], {'max_length': '344'}),
            'hitehre': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['foobar.Baz']", 'unique': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'situp': ('django.db.models.fields.IntegerField', [], {})
        }
    }

    complete_apps = ['foobar']



from django.core.urlresolvers import reverse
from django.test import TestCase

from easymode.tests.testcases import initdb


__all__ = ('TestL10nForm',)

@initdb
class TestL10nForm(TestCase):
    """Tests for the introspection package in easymode"""
    
    fixtures = ['auth-group', 'auth-user', 'formtestmodel']
    
    post_data = {
        'number': 130,
        'other': 'Your mother',
        'another': 'Well well',
    }
    
    username = 'admin'
    password = 'admin'
    
    def test_form_is_overridden(self):
        """The form on the admin should be overridden to have a max value for number"""
        self.assertTrue(self.client.login(username=self.username, password=self.password))
        admin_url = reverse('admin:tests_formtestmodel_change', args=[1])
        response = self.client.get(admin_url)
        self.failUnlessEqual(response.status_code, 200)
        
        # test overridden internationalized field
        self.assertContains(response, 100)
        # test normal field
        self.assertContains(response, 'hello')
        # test normal internationalized field
        self.assertContains(response, 'again')
        
        response = self.client.post(admin_url, self.post_data)
        self.assertContains(response, 130)
        self.assertContains(response, 'Your mother')
        self.assertContains(response, 'Well well')
        
        # the overridden number field should have set an error
        errors = response.context['errors']
        self.assertEquals(len(errors), 1)
        self.assertEquals(len(errors[0]), 1)
        self.assertEquals( errors[0][0], u'Ensure this value is less than or equal to 100.')



"""
Derived from django-transmeta.
Contains machinery to change a type after construction,
by using a class decorator.

This in contrast with django-transmeta which requires
overriding the __metaclass__.
"""
import copy

from django.conf import settings
from django.db.models.fields import NOT_PROVIDED
from django.contrib.admin.options import FORMFIELD_FOR_DBFIELD_DEFAULTS
from django.utils import translation
from django.utils.encoding import force_unicode
from django.utils.translation.trans_real import translation as translation_catalogs

from easymode.utils import first_match
from easymode.utils.standin import standin_for
from easymode.utils.languagecode import get_all_language_codes, get_real_fieldname

__all__ = ('GettextVO',)

def valid_for_gettext(value):
    """Gettext acts weird when empty string is passes, and passing none would be even weirder"""
    return value not in (None, "")
    
def get_fallback_languages():
    """Retrieve the fallback languages from the settings.py"""
    lang = translation.get_language()
    fallback_list = settings.FALLBACK_LANGUAGES.get(lang, None)
    if fallback_list:
        return fallback_list

    return settings.FALLBACK_LANGUAGES.get(lang[:2], [])


def get_localized_property(context, field=None, language=None):
    '''
    When accessing to the name of the field itself, the value
    in the current language will be returned. Unless it's set,
    the value in the default language will be returned.
    '''
    if language:
        return getattr(context, get_real_fieldname(field, language))
    
    if hasattr(settings, 'FALLBACK_LANGUAGES'):
        attrs = [translation.get_language()]
        attrs += get_fallback_languages()
    else:
        attrs = [
            translation.get_language(),
            translation.get_language()[:2],
            settings.LANGUAGE_CODE, 
        ]
    
    def predicate(x):
        value = getattr(context, get_real_fieldname(field, x), None)
        return value if valid_for_gettext(value) else None

    return first_match(predicate, attrs)


def get_localized_field_name(context, field):
    """Get the name of the localized field"""
    attrs = [
             translation.get_language(), 
             translation.get_language()[:2], 
             settings.LANGUAGE_CODE
            ]
            
    def predicate(x):
        field_name = get_real_fieldname(field, x)
        if hasattr(context, field_name):
            return field_name
        return None

    return first_match(predicate, attrs)


class GettextVO:
    """
    A value object that contains information about the origin
    of the value in question.
    
    .. attribute:: standin_value_is_from_database

        is True if the value came from the database

    .. attribute:: msgid
    
        The message id that originated the lookup

    .. attribute:: msg
        
        The message that is the result of the lookup
    
    .. attribute:: fallback
    
        the fallback that was found during the lookup
    
    .. attribute:: stored_value
        
        the value as it is stored in the database
    
    """
    
    def __init__(self):
        self.standin_value_is_from_database = False
        self.msgid = None
        self.msg = None
        self.fallback = None
        self.stored_value = None


class DefaultFieldDescriptor(property):
    """
    Descriptor that implements access to the default language.
    It inherits property because django wants that, it doesn't process
    it as a field if it is not isinstance(self, property)
    """

    def __init__(self, name, **kwargs):
        """The name param is the name of the field the descriptor is emulating."""
        self.name = name
        self.__dict__.update(kwargs)

    def __get__(self, obj, typ=None):
        """
        Read the localised version of the field this descriptor emulates.
        First try to see if the localised field is really set.
        If not, then use ugettext_lazy to find a tranlation in the current language
        for this field.
        """
        # self must be returned in a getattr context.
        if obj is None:
            return self

        current_language = translation.get_language()
        real_field_name = get_real_fieldname(self.name, current_language)

        vo = GettextVO()
        
        # first check if the database contains the localized data
        vo.stored_value = getattr(obj, real_field_name)

        # the database does not have our localized data.
        # check if we have a translation, first get the msgid, as a unicode string.
        vo.msgid = get_localized_property(obj, self.name, getattr(settings, 'MSGID_LANGUAGE', settings.LANGUAGE_CODE))

        # check the translation in the current language
        # but avoid empty string and None 
        if valid_for_gettext(vo.msgid):
            vo.msg = self.to_python(translation.ugettext(force_unicode(vo.msgid)))
        elif valid_for_gettext(vo.stored_value):
            # we can not use the msgid for gettext but we did find a valid
            # translation in the database. Fine we stop here and return that
            # value. No need for a standin, because we don't have a catalog value.
            return vo.stored_value
        else:
            # we can not use the msgid for gettext lookups, so there is no
            # point in trying. Check for fallback languages in database.
            vo.fallback = get_localized_property(obj, self.name)
            
            if not valid_for_gettext(vo.fallback):
                # Also if we are sure we don't have any old
                # translations in the catalog or something for the fallback
                # languages in the database, we do not need to return a standin either
                return vo.msgid
        
        # we got here so we've got a valid messageid. Now collect data from the catalog(s)
        
        # if there isn't anything new in the catalog belonging to the current language:
        if vo.msg == vo.msgid:
            # maybe we have a translation in any of the fallback languages.
            if hasattr(settings, 'FALLBACK_LANGUAGES'):
                # first check if the database has the localized data in
                # any of the fallback languages.
                vo.fallback = get_localized_property(obj, self.name)
                
                # if the fallback is the same as the msgid, go and look in the catalog
                if vo.fallback == vo.msgid:
                    # there might be a translation in any
                    # of the fallback languages.
                    for fallback in get_fallback_languages():
                        catalog = translation_catalogs(fallback)
                        msg = catalog.ugettext(force_unicode(vo.msgid))
                        if self.to_python(msg) != vo.msgid:
                            vo.fallback = self.to_python(msg)
                            break
                elif vo.fallback:
                    # if a valid fallback is found, then, since the msg is equal
                    # to the msgid, the fallback is the winner.
                    vo.msg = vo.fallback

        # if we came here we collected data from the catalog and we should return
        # a standin. A standin is the return value, with some extra properties.
        # see GettextVO for the extra properties added.
        if valid_for_gettext(vo.stored_value):
            vo.standin_value_is_from_database = True
            # database always wins
            return standin_for(vo.stored_value, **vo.__dict__)
        elif valid_for_gettext(vo.msg):
            # runner up is the translation in the native language
            return standin_for(vo.msg, **vo.__dict__)
        elif valid_for_gettext(vo.fallback):
            # and last is the translation in a fallback language
            return standin_for(vo.fallback, **vo.__dict__)

        assert(valid_for_gettext(vo.msgid))

        # there is a very very small probability that the translation of
        # a valid msgid evaluates to empty string or None (after to_python).
        # If that happened, we got here. I choose to return None, because i like
        # to be like that
        return None

    def __set__(self, obj, value):
        """Write the localised version of the field this descriptor emulates."""
        local_property = get_localized_field_name(obj, self.name)
        setattr(obj, local_property, value)

    def get_internal_type(self):
        """Used to describe the type while serializing."""
        return "CharField"

    def value_to_string(self, obj):
        """This descriptor acts as a Field, as far as the serializer is concerned."""
        try:
            return force_unicode(self.__get__(obj))
        except TypeError:
            return str(self.__get__(obj))


def localize_fields(cls, localized_fields):
    """
    For each field name in localized_fields,
    for each language in settings.LANGUAGES,
    add fields to cls,
    and remove the original field, instead
    replace it with a DefaultFieldDescriptor,
    which always returns the field in the current language.
    """
    
    # never do this twice
    if hasattr(cls, 'localized_fields'):
        return cls
        
    # MSGID_LANGUAGE is the language that is used for the gettext message id's.
    # If it is not available, because the site isn't using subsites, the LANGUAGE_CODE
    # is good too. MSGID_LANGUAGE gives the opportunity to specify a language not available
    # in the site but which is still used for the message id's.
    msgid_language = getattr(settings, 'MSGID_LANGUAGE',  settings.LANGUAGE_CODE)
    
    # set the localized fields property
    cls.localized_fields = localized_fields
    
    for field in localized_fields:
        original_attr = cls._meta.get_field_by_name(field)[0]
        
        for cnt, language_code in enumerate(get_all_language_codes()):
            i18n_attr = copy.copy(original_attr)
            # add support for south introspection.
            i18n_attr._south_introspects = True
            i18n_attr.original_fieldname = field
            i18n_attr.include_in_xml = False
            lang_attr_name = get_real_fieldname(field, language_code)
            i18n_attr.creation_counter = i18n_attr.creation_counter + .01 * cnt
            # null must be allowed for the message id language because this language
            # might not be available at all in the backend
            if not i18n_attr.null and i18n_attr.default is NOT_PROVIDED:
                i18n_attr.null = True
            
            if language_code != msgid_language:
                # no validation for the fields that are language specific
                if not i18n_attr.blank:
                    i18n_attr.blank = True
                    
            if i18n_attr.verbose_name:
                i18n_attr.verbose_name = translation.string_concat(i18n_attr.verbose_name, u' (%s)' % language_code)
            cls.add_to_class(lang_attr_name, i18n_attr)
            
        # delete original field
        del cls._meta.local_fields[cls._meta.local_fields.index(original_attr)]

        # copy some values and functions from the original_attr
        # so the field can emulate the original_attr as good as possible
        kwargs = {
            'serialize': getattr(original_attr, 'serialize', True),
            'extra_attrs':getattr(original_attr, 'extra_attrs', None),
            'max_length': getattr(original_attr, 'max_length', None),
            'min_length' : getattr(original_attr, 'min_length', None),
            'form_field' : original_attr.formfield(**FORMFIELD_FOR_DBFIELD_DEFAULTS.get(original_attr.__class__, {})),
            'get_internal_type': original_attr.get_internal_type,
            'unique': getattr(original_attr, 'unique', False),
            'to_python': original_attr.to_python,
        }
            
        # copy custom_value_serializer if it was defined on the original attr
        if hasattr(original_attr, 'custom_value_serializer'):
            kwargs['custom_value_serializer'] = original_attr.custom_value_serializer

        # add the DefaultFieldDescriptor where the original_attr was.
        cls.add_to_class(field, DefaultFieldDescriptor(field, **kwargs))
        
        # update fields cache
        cls._meta._fill_fields_cache()

    # return the finished product
    return cls



from django import http
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist

from easymode.utils.languagecode import fix_language_code

@staff_member_required
def preview(request, content_type_id, object_id):
    """
    This is an override for django.views.default.shortcut.
    It assumes that get_absolute_url returns an absolute url, so
    it does not do any of the very elaborate link checking that shortcut does.
    
    This version adds the language code to the url. (/en/blaat/).
    """
    
    try:
        content_type = ContentType.objects.get(pk=content_type_id)
        obj = content_type.get_object_for_this_type(pk=object_id)
    except ObjectDoesNotExist:
        raise http.Http404("Content type %s object %s doesn't exist" % (content_type_id, object_id))
    try:
        absolute_url = obj.get_absolute_url(format='flash')
    except AttributeError:
        raise http.Http404("%s objects don't have get_absolute_url() methods" % content_type.name)
        
    if absolute_url.startswith('http://') or absolute_url.startswith('https://'):        
        http.HttpResponseRedirect(absolute_url)        
    else:
        absolute_url = fix_language_code("/%s" % absolute_url, request.LANGUAGE_CODE)                
        return http.HttpResponseRedirect(absolute_url)


"""
    easy_locale <targetdir> <applabel>
    
    Will create a folder locale in targetdir with locales who's
    data comes from the models in applabel.
    
    example:
    
    ./manage.py easy_locale myapp myapp
    
    will create myapp/locale/ with po files in it.
    
    Or:
    ./manage.py easy_locale myapp myapp.mymodel
    
    which will only dump the MyModel objects to the database.
"""

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import get_models, get_app, get_model
from django.utils import translation

from easymode.i18n.gettext import MakeModelMessages

class Command(BaseCommand):
    
    help = __doc__
    
    def handle(self, targetdir, app=None, **options):
        """ command execution """
        translation.activate(settings.LANGUAGE_CODE)
        if app:
            unpack = app.split('.')

            if len(unpack) == 2:
                models = [get_model(unpack[0], unpack[1])]
            elif len(unpack) == 1:
                models = get_models(get_app(unpack[0]))
        else:
            models = get_models()
        
        messagemaker = MakeModelMessages(targetdir)
        
        for model in models:
            if hasattr(model, 'localized_fields'):
                for instance in model.objects.all():
                    messagemaker(instance)